{ Ager's Socket Library (c) Copyright 1998-00 by Soren Ager (sag@poboxes.com) }
{                                                                             }
{ $Revision: 1.22 $    $Date: 2000/08/24 01:42:28 $    $Author: sag $ }
{                                                                             }
{ Abstract Client class                                                       }
UNIT aslAbsClient;

{$I aslDefine.Inc}

INTERFACE

USES Classes, SysUtils, aslAbsSocket, aslTCPSocket;

TYPE
  EAbsClient = class(EaslException);

  TAbsClient = CLASS
  PRIVATE
    FHost             : STRING;
    FService          : STRING;
    FGreeting         : STRING;
    FLastResponse     : STRING;
    FLastResponseCode : Word;
    FOnLogLine        : TOnLogLineProc;

  PROTECTED
    FSocket           : TTCPClientSocket;

    PROCEDURE SendCommand(Cmd: STRING);
    PROCEDURE SendCommandEx(Cmd: STRING; Extra: TStringList);
    PROCEDURE GetResponse(Extra: TStringList); VIRTUAL;
    PROCEDURE GetMsgLines(StringList: TStringList);
    PROCEDURE SendMsgLines(StringList: TStringList);
    PROCEDURE GetLines(StringList: TStringList);
    PROCEDURE SendLines(StringList: TStringList);
    PROCEDURE LogLine(Msg: STRING);

  PUBLIC
    CONSTRUCTOR Create;
    DESTRUCTOR Destroy; OVERRIDE;
    PROCEDURE Connect(Host: STRING); VIRTUAL;
    PROCEDURE Disconnect; VIRTUAL;

    PROPERTY Socket: TTCPClientSocket      READ FSocket;
    PROPERTY Host: STRING                  READ FHost;
    PROPERTY Service: STRING               READ FService          WRITE FService;
    PROPERTY Greeting: STRING              READ FGreeting;
    PROPERTY LastResponse: STRING          READ FLastResponse     WRITE FLastResponse;
    PROPERTY LastResponseCode: Word        READ FLastResponseCode WRITE FLastResponseCode;
    PROPERTY OnLogLine: TOnLogLineProc     READ FOnLogLine        WRITE FOnLogLine;
  END;

IMPLEMENTATION

  CONSTRUCTOR TAbsClient.Create;
  BEGIN
    INHERITED Create;
    FSocket:=TBufTCPClientSocket.Create;
    FHost:='';
    FService:='';
    FGreeting:='';
    FLastResponse:='';
    FLastResponseCode:=0;
    FOnLogLine:=NIL;
  END;

  DESTRUCTOR TAbsClient.Destroy;
  BEGIN
    IF Assigned(FSocket) THEN
    BEGIN
      FSocket.Destroy;
      FSocket:=Nil;
    END;
    INHERITED Destroy;
  END;

  PROCEDURE TAbsClient.Connect(Host: STRING);
  BEGIN
    IF FService='' THEN
      RAISE EAbsClient.Create('Connect: No service specified', '');
    IF NOT Assigned(FSocket) THEN
      RAISE EAbsClient.Create('Connect: Socket is not initialized (Service='+FService+')', '');
    FHost:=Host;
    FSocket.Connect(FHost, FService);
    GetResponse(Nil);
    FGreeting:=FLastResponse;
  END;

  PROCEDURE TAbsClient.Disconnect;
  BEGIN
    FSocket.Disconnect;
    FHost:='';
    FService:='';
    FLastResponse:='';
    FLastResponseCode:=0;
  END;


  PROCEDURE TAbsClient.SendCommand(Cmd: STRING);
  BEGIN
    LogLine('<'+Cmd);
    FSocket.WriteLn(Cmd);
    GetResponse(Nil);
  END;

  PROCEDURE TAbsClient.SendCommandEx(Cmd: STRING; Extra: TStringList);
  BEGIN
    LogLine('<'+Cmd);
    FSocket.WriteLn(Cmd);
    GetResponse(Extra);
  END;

  PROCEDURE TAbsClient.GetResponse(Extra: TStringList);
  VAR
    Ok : Integer;
    S  : String;
  BEGIN
    FSocket.ReadLn(FLastResponse);
    LogLine('>'+FLastResponse);
    Val(Copy(FLastResponse,1,3), FLastResponseCode, Ok);

    IF Copy(FLastResponse,4,1)='-' THEN
    BEGIN
      REPEAT
        FSocket.ReadLn(S);
        IF Extra<>nil THEN Extra.Add(S);
        LogLine('>'+S);
      UNTIL Copy(S,1,4)=(Copy(FLastResponse,1,3)+' ');
    END;

{
    WHILE Copy(FLastResponse,4,1)='-' DO
    BEGIN
      FSocket.ReadLn(FLastResponse);
      IF Extra<>nil THEN Extra.Add(Copy(FLastResponse,5,Length(FLastResponse)-4));
      LogLine('>'+FLastResponse);
    END;
}

  END;

  PROCEDURE TAbsClient.GetMsgLines(StringList: TStringList);
  VAR
    S, S1 : STRING;
  BEGIN
    REPEAT
      FSocket.ReadLn(S);
      S1:=S;
      IF (Length(S)>1) AND (Copy(S,1,1)='.') THEN S:=Copy(S, 2, Length(S)-1);
      IF (S1<>'.') AND Assigned(StringList) THEN StringList.Add(S);
    UNTIL S1='.';
  END;

  PROCEDURE TAbsClient.SendMsgLines(StringList: TStringList);
  VAR
    i : Integer;
  BEGIN
    IF Assigned(StringList) THEN
      FOR i:=0 TO StringList.Count-1 DO
        IF (StringList[i]<>'') AND (Copy(StringList[i],1,1)='.') THEN
          Socket.WriteLn('.'+StringList[i])
        ELSE
          Socket.WriteLn(StringList[i]);
  END;

  PROCEDURE TAbsClient.GetLines(StringList: TStringList);
  VAR
    S : STRING;
  BEGIN
    WHILE FSocket.WaitForDataTime(1000)>0 DO
    BEGIN
      FSocket.ReadLn(S);
      LogLine('>'+S);
      IF Assigned(StringList) THEN StringList.Add(S);
    END;
  END;

  PROCEDURE TAbsClient.SendLines(StringList: TStringList);
  VAR
    i : Integer;
  BEGIN
    IF Assigned(StringList) THEN
      FOR i:=0 TO StringList.Count-1 DO
      Socket.WriteLn(StringList[i]);
  END;

  PROCEDURE TAbsClient.LogLine(Msg: STRING);
  BEGIN
    IF Assigned(FOnLogLine) THEN FOnLogLine(Msg);
  END;

END.


